package cmdTemplate

import (
	"errors"
	"fmt"
	"regexp"
	"strings"

	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"

	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"

	t "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/template"

	"github.com/google/uuid"
)

type List struct {
	Name string
}

func (cmd *List) GetName() string {
	return cmd.Name
}

func (cmd *List) GetDesc() []string {
	return []string{
		"Lists templates.",
		"Requires TPL_LIST or TPL_LIST_ANY user capability."}
}

func (cmd *List) PrintUsage() {
	for _, desc := range cmd.GetDesc() {
		u.PrintlnErr(desc)
	}
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("USAGE: ", cmd.GetName(), " [-l] [ID ...] [regex ...]")
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("Use \"-l\" to specify detailed templates output.")
	u.PrintlnErr("The action only applies to templates matching the specified IDs or regexes.")
	const usage string = `Any number of IDs or regexes can be specified.
The regex only matches the template's name and is case-insensitive.
Regex examples:
"."   -> matches any templates
"bla" -> matches any templates containing "bla" in their name`
	u.PrintlnErr(usage)
}

// Prints a list of filtered templates.
// Return 0 if everything went well or 1 in case of failure.
func (cmd *List) Run(nc *nc.NexusClient, args []string) int {
	if len(args) < 1 {
		cmd.PrintUsage()
		return 1
	}

	// Check if a "-l" argument is specified
	foundLongOutputFlag := -1
	for idx, arg := range args {
		if arg == "-l" {
			foundLongOutputFlag = idx
		}
	}

	if foundLongOutputFlag >= 0 {
		args = utils.RemoveArgAtIndex(args, foundLongOutputFlag)
	}

	templates, err := cmd.getFilteredTemplates(nc, args)
	if err != nil {
		u.PrintlnErr(err)
		return 1
	}

	if foundLongOutputFlag >= 0 {
		for _, template := range templates {
			str, err := template.String()
			if err != nil {
				u.PrintlnErr("Failed decoding template " + template.ID.String() + ". Template skipped.")
			} else {
				u.Println(str)
			}
		}
	} else {
		// Compute the length of the longest name and access (used for padding columns)
		nameLen := 0
		idLen := 0
		accessLen := 0
		for _, template := range templates {
			l := len(template.Name)
			if l > nameLen {
				nameLen = l
			}
			idLen = len(template.ID.String()) // ID is a UUID of constant length
			l = len(template.Access)
			if l > accessLen {
				accessLen = l
			}
		}

		paddedFmt := fmt.Sprintf("%%-%ds | %%-%ds | %%-%ds\n", nameLen, idLen, accessLen)

		for _, template := range templates {
			fmt.Printf(paddedFmt, template.Name, template.ID, template.Access)
		}
	}

	return 0
}

// Returns a list of filtered templates for a given route.
// Filters are based on patterns describing IDs or regexes.
// Patterns can either be:
// - any number of "template IDs" (UUID)
// - any number of "template names"
// - any combination of "template IDs" and "template names"
// - any regular expression (it applies to the template's name only)
// Remark: the matching is case-insensitive
// Regular expression examples:
//
//	"."   -> matches everything
//	"bla" -> matches any template name containing "bla"
func (cmd *List) getFilteredTemplates(nc *nc.NexusClient, patterns []string) ([]t.TemplateSerialized, error) {
	if len(patterns) < 1 {
		return nil, errors.New("At least one ID or regex must be specified")
	}

	var ids []string
	var regexes []string

	for _, pattern := range patterns {
		_, err := uuid.Parse(pattern)
		if err != nil {
			regexes = append(regexes, pattern)
		} else {
			ids = append(ids, pattern)
		}
	}

	templates, err := nc.GetTemplates()
	if err != nil {
		return nil, err
	}

	filteredTemplates := []t.TemplateSerialized{}

	for _, template := range templates {
		found := false
		for _, id := range ids {
			if id == template.ID.String() {
				filteredTemplates = append(filteredTemplates, template)
				found = true
				break
			}
		}
		if found {
			continue
		}
		for _, regex := range regexes {
			match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(template.Name))
			if err != nil {
				return nil, errors.New("Error matching \"" + regex + "\": " + err.Error())
			} else {
				if match {
					filteredTemplates = append(filteredTemplates, template)
					break
				}
			}
		}
	}
	return filteredTemplates, nil
}
