package cmdUser

import (
	"fmt"
	"regexp"
	"strings"

	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"

	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
)

type List struct {
	Name string
}

func (cmd *List) GetName() string {
	return cmd.Name
}

func (cmd *List) GetDesc() []string {
	return []string{
		"Lists users.",
		"Requires USER_LIST user capability."}
}

func (cmd *List) PrintUsage() {
	for _, desc := range cmd.GetDesc() {
		u.PrintlnErr(desc)
	}
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("USAGE: ", cmd.GetName(), " [-l] [regex ...]")
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("Use \"-l\" to specify detailed user output.")
	printRegexUsageDetails()
}

// Prints a list of filtered users for a given route.
// The filter is based on regexes.
// Remark: the regex matches the user email, first name and last name and is case-insensitive.
// Regular expression examples:
//
//	""    -> matches everything
//	"."   -> matches everything
//	"bla" -> matches any user containing "bla"
func (cmd *List) Run(nc *nc.NexusClient, args []string) int {
	if len(args) < 1 {
		cmd.PrintUsage()
		return 1
	}

	// Check if a "-l" argument is specified
	foundLongOutputFlag := -1
	for idx, arg := range args {
		if arg == "-l" {
			foundLongOutputFlag = idx
		}
	}

	if foundLongOutputFlag >= 0 {
		args = utils.RemoveArgAtIndex(args, foundLongOutputFlag)
	}

	users, err := nc.GetUsers()
	if err != nil {
		u.PrintlnErr(err)
		return 1
	}

	userList := []params.UserWithoutPwd{}

	for _, user := range users {
		for _, regex := range args {
			fieldsToMatch := user.Email + " " + user.FirstName + " " + user.LastName
			match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(fieldsToMatch))
			if err != nil {
				u.PrintlnErr("Error matching \"" + regex + "\": " + err.Error())
			} else {
				if match {
					userList = append(userList, user)
					break
				}
			}
		}
	}

	if len(userList) == 0 {
		u.PrintlnErr("Error: user(s) not found.")
		return 1
	}

	if foundLongOutputFlag >= 0 {
		for _, user := range userList {
			u.Println(user.String())
		}
	} else {
		// Compute the length of the longest LastName, FirstName and Email (used for padding columns)
		lastNameLen := 0
		firstNameLen := 0
		emailLen := 0
		for _, user := range userList {
			l := len(user.LastName)
			if l > lastNameLen {
				lastNameLen = l
			}
			l = len(user.FirstName)
			if l > firstNameLen {
				firstNameLen = l
			}
			l = len(user.Email)
			if l > emailLen {
				emailLen = l
			}
		}

		paddedFmt := fmt.Sprintf("%%-%ds | %%-%ds | %%-%ds\n", lastNameLen, firstNameLen, emailLen)

		for _, user := range userList {
			fmt.Printf(paddedFmt, user.LastName, user.FirstName, user.Email)
		}
	}

	return 0
}

func printRegexUsageDetails() {
	const usage string = `Only users matching the specified regexes will be listed.
Any number of regexes can be specified.
The regex matches the user email, first name and last name and is case-insensitive.
Regex examples:
"."   -> matches any users
"bla" -> matches any users containing "bla"`
	u.PrintlnErr(usage)
}
