package utils

import (
	"archive/tar"
	"bytes"
	"compress/gzip"
	"errors"
	"io"
	"io/fs"
	"os"
	"path/filepath"
	"strings"

	"github.com/google/uuid"
)

// Returns true if the specified file exists, false otherwise.
func FileExists(filename string) bool {
	_, err := os.Stat(filename)
	return !errors.Is(err, os.ErrNotExist)
}

func GetRandomTempFilename() (string, error) {
	tempDir := os.TempDir()
	uuid, err := uuid.NewRandom()
	if err != nil {
		return "", errors.New("Failed creating random UUID: " + err.Error())
	}
	randName := "temp_" + uuid.String()
	return filepath.Join(tempDir, randName), nil
}

func CreateRandomEmptyFile() (string, error) {
	filePath, err := GetRandomTempFilename()
	if err != nil {
		return "", err
	}

	file, err := os.Create(filePath)
	if err != nil {
		return "", err
	}
	file.Close()

	return filePath, nil
}

// Creates a tar.gz archive of dir and all its files and subdirectories.
// Note: dir can also be a file.
// Source code slightly modified from: https://gist.github.com/mimoo/25fc9716e0f1353791f5908f94d6e726
func TarGzDir(dir, archive string) error {
	_, err := os.Stat(dir)
	if errors.Is(err, fs.ErrNotExist) {
		return errors.New("Error: \"" + dir + "\" does not exist")
	}

	var buf bytes.Buffer

	gzWriter := gzip.NewWriter(&buf)
	tarWriter := tar.NewWriter(gzWriter)

	// Walks through every file in the directory
	filepath.Walk(dir, func(file string, fi os.FileInfo, e error) error {
		header, err := tar.FileInfoHeader(fi, file)
		if err != nil {
			return err
		}

		fileRel := strings.TrimPrefix(file, dir)
		fileRel = filepath.Join(filepath.Base(dir), fileRel)
		header.Name = filepath.ToSlash(fileRel)

		if err := tarWriter.WriteHeader(header); err != nil {
			return err
		}
		// if not a dir, write file content
		if !fi.IsDir() {
			data, err := os.Open(file)
			if err != nil {
				return err
			}
			if _, err := io.Copy(tarWriter, data); err != nil {
				return err
			}
		}
		return nil
	})

	if err := tarWriter.Close(); err != nil {
		return err
	}
	if err := gzWriter.Close(); err != nil {
		return err
	}

	fileToWrite, err := os.OpenFile(archive, os.O_CREATE|os.O_RDWR, os.FileMode(0750))
	if err != nil {
		return err
	}
	if _, err := io.Copy(fileToWrite, &buf); err != nil {
		return err
	}

	return nil
}

// Returns the list of subdirectories present in dir.
func GetSubDirs(dir string) ([]string, error) {
	subDirs := []string{}

	currentDir, err := os.Open(dir)
	if err != nil {
		return nil, err
	}

	// Retrieves all files entries in the directory (0 = all files in the directory).
	files, err := currentDir.ReadDir(0)
	if err != nil {
		currentDir.Close()
		return nil, err
	}

	currentDir.Close()

	// Loop over file entries
	for _, f := range files {
		info, _ := f.Info()
		if f.IsDir() || info.Mode()&os.ModeSymlink != 0 {
			subDirs = append(subDirs, filepath.Join(dir, f.Name()))
		}
	}

	return subDirs, nil
}

func CopyFiles(source, dest string) error {
	src, err := os.Open(source)
	if err != nil {
		return err
	}
	defer src.Close()

	dst, err := os.Create(dest)
	if err != nil {
		return err
	}
	defer dst.Close()
	_, err = io.Copy(dst, src)

	return err
}
