package nexusclient

import (
	"crypto/tls"
	"encoding/json"
	"errors"
	"net/url"
	"os"
	"path/filepath"
	"time"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/response"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"
	"github.com/go-resty/resty/v2"
	"github.com/google/uuid"
)

type NexusClient struct {
	client   *resty.Client
	hostname string // host, e.g. "127.0.0.1" or "my.domain.net"
	certFile string // path to the CA certificate tied to the hostname above
	// (this cert is required by the spice client)
}

// The host argument has the following format:
// "ip:port" or "domain:port", e.g. "127.0.0.1:1077" or "my.domain.net:80"
func New(host string) (*NexusClient, error) {
	nc := &NexusClient{}
	nc.client = resty.New()

	// To disable the check of the CA signed certificate
	nc.client.SetTLSClientConfig(&tls.Config{InsecureSkipVerify: true})
	nc.client.SetBaseURL("https://" + host)
	nc.client.SetAllowGetMethodPayload(true)
	// Change the default timeout to 10 seconds
	nc.client.SetTimeout(10 * time.Second)

	url, err := url.Parse(nc.client.BaseURL)
	if err != nil {
		return nil, err
	}
	nc.hostname = url.Hostname()
	nc.certFile = ""
	return nc, nil
}

// Authenticate a user.
// curl --cacert ca-cert.pem -X POST https://localhost:8000/login -H 'Content-Type: application/json' -d '{"email": "johndoe@nexus.org", "pwd":"pipomolo"}'
// Returns an JWT token if authentication succeeded or an error if it failed.
// IMPORTANT: caller MUST call the Cleanup() function before exiting!
func (nc *NexusClient) Authenticate(user, pwd string) (string, error) {
	loginArgs := &params.Login{Email: user, Pwd: pwd}
	resp, err := nc.client.R().SetBody(loginArgs).Post("/login")
	if err != nil {
		return "", err
	}

	if resp.IsSuccess() {
		var token params.Token
		err = json.Unmarshal(resp.Body(), &token)
		if err != nil {
			return "", err
		}
		nc.client.SetAuthToken(token.Token)
		// Download the CA certificate required by Spice clients
		nc.certFile, err = nc.downloadVMAttachCert()
		if err != nil {
			return "", err
		}
		return token.Token, nil
	} else {
		return "", response.ErrorToMsg(resp)
	}
}

// "Authenticate" by using an already established connection using a previously received token.
// Useful when wanting to access the API without authenticating.
// IMPORTANT: caller MUST call the Cleanup() function before exiting!
func (nc *NexusClient) AuthenticateWithToken(token string) error {
	nc.client.SetAuthToken(token)
	// Download the CA certificate required by Spice clients
	var err error
	nc.certFile, err = nc.downloadVMAttachCert()
	if err != nil {
		return err
	}
	return nil
}

// Cleanup resources allocated by Authenticate and AuthenticateWithToken.
// IMPORTANT: caller MUST call this function before exiting!
func (nc *NexusClient) Cleanup() {
	if nc.certFile != "" {
		os.Remove(nc.certFile)
	}
}

// Returns the hostname, e.g. "127.0.0.1" or "my.domain.net".
func (nc *NexusClient) GetHostname() string {
	return nc.hostname
}

// Change the default timeout.
func (nc *NexusClient) SetTimeout(timeout time.Duration) {
	nc.client.SetTimeout(timeout)
}

// Obtain a new JWT token from the server.
func (nc *NexusClient) RefreshToken() error {
	resp, err := nc.client.R().Get("/misc/token")
	if err != nil {
		return err
	} else {
		if resp.IsSuccess() {
			var token params.Token
			err = json.Unmarshal(resp.Body(), &token)
			if err != nil {
				return err
			}
			nc.client.SetAuthToken(token.Token)
			return nil
		} else {
			return response.ErrorToMsg(resp)
		}
	}
}

func (nc *NexusClient) getVMAttachCertFile() string {
	return nc.certFile
}

// Retrieve and save the certificate required by the various VMAttachXXX methods.
// This public CA certificate is required by Spice remote viewers.
// The certificate file is saved in the OS' temporary directory.
// Its name is randomly generated and garanteed to be unique.
func (nc *NexusClient) downloadVMAttachCert() (string, error) {
	nc.Cleanup()

	uniqPostfix, err := uuid.NewRandom()
	if err != nil {
		return "", errors.New("Failed saving CA certificate, code 1:\n" + err.Error())
	}

	outputFile := filepath.Join(os.TempDir(), "cert_"+uniqPostfix.String()+".pem")
	resp, err := nc.client.R().SetOutput(outputFile).Get("/misc/cert")
	if err != nil {
		return "", errors.New("Failed saving CA certificate, code 2:\n" + err.Error())
	}

	if resp.IsSuccess() {
		return outputFile, nil
	} else {
		errorMsg, _ := u.FileToString(outputFile)
		return "", errors.New("Failed saving CA certificate, code 3\n" + resp.Status() + ": " + errorMsg)
	}
}
