package nexusclient

import (
	"encoding/json"
	"errors"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/vm"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/response"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"
	"github.com/go-playground/validator/v10"
	"github.com/go-resty/resty/v2"
)

const (
	attachVMs     = "/vms/attach"
	delVMs        = "/vms/del"
	editVMs       = "/vms/edit"
	editaccessVMs = "/vms/editaccess"
	exportdirVMs  = "/vms/exportdir"
	importfileVMs = "/vms/importfiles"
	listVMs       = "/vms"
	rebootVMs     = "/vms/reboot"
	startVMs      = "/vms/start"
	stopVMs       = "/vms/stop"
)

//========================================================================================
// Public functions
//========================================================================================

func (nc *NexusClient) VMCreate(p params.VMCreate) (*vm.VMNetworkSerialized, error) {
	resp, err := nc.client.R().SetBody(p).Post("/vms")
	if err != nil {
		return nil, err
	}
	if resp.IsSuccess() {
		var vm vm.VMNetworkSerialized
		if err := json.Unmarshal(resp.Body(), &vm); err != nil {
			return nil, err
		}
		return &vm, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMEdit(vmID string, p params.VMEdit) (*vm.VMNetworkSerialized, error) {
	resp, err := nc.client.R().SetBody(p).Put("/vms/" + vmID)
	if err != nil {
		return nil, err
	}
	if resp.IsSuccess() {
		var vm vm.VMNetworkSerialized
		if err := json.Unmarshal(resp.Body(), &vm); err != nil {
			return nil, err
		}
		return &vm, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMDel(vmID string) error {
	resp, err := nc.client.R().Delete("/vms/" + vmID)
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMExportDir(vmName, vmID, dirToExport string) (string, error) {
	p := &params.VMExportDir{Dir: dirToExport}

	outputFile := vmName + "_" + vmID + ".tar.gz"
	resp, err := nc.client.R().SetOutput(outputFile).SetBody(p).Get("/vms/" + vmID + "/exportdir")
	if err != nil {
		// u.PrintlnErr("Failed exporting " + dir + " from VM \"" + vm.Name + "\": " + err.Error())
		return "", err
	}
	if resp.IsSuccess() {
		return outputFile, nil
	} else {
		errorMsg, _ := u.FileToString(outputFile)
		return "", errors.New("Failed exporting " + dirToExport + " from VM \"" + vmName + "\": " + resp.Status() + ": " + errorMsg)
	}
}

func (nc *NexusClient) VMDelAccess(vmID, vmName, email string) error {
	resp, err := nc.client.R().Delete("/vms/" + vmID + "/access/" + email)
	if err != nil {
		return err
	}

	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMImportArchive(vmID, vmDir, archiveFile string) error {
	resp, err := nc.client.R().SetFile("file", archiveFile).
		SetFormData(map[string]string{
			"vmDir": vmDir,
		}).Post("/vms/" + vmID + "/importfiles")

	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMReboot(vmID string) error {
	resp, err := nc.client.R().Put("/vms/" + vmID + "/reboot")
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMShutdown(vmID string) error {
	resp, err := nc.client.R().Put("/vms/" + vmID + "/shutdown")
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMStart(vmID string) error {
	resp, err := nc.client.R().Put("/vms/" + vmID + "/start")
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMStartWithCreds(vmID string, p params.VMStartWithCreds) error {
	resp, err := nc.client.R().SetBody(p).Put("/vms/" + vmID + "/startwithcreds")
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMStop(vmID string) error {
	resp, err := nc.client.R().Put("/vms/" + vmID + "/stop")
	if err != nil {
		return err
	}
	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) VMAttachFromID(vmID string, fullscreen bool) error {
	creds, err := nc.vmIdToAttachCreds(vmID)
	if err != nil {
		return err
	}

	err = nc.VMAttachFromCreds(*creds, fullscreen)
	if err != nil {
		return err
	}
	return nil
}

func (nc *NexusClient) VMAttachFromCreds(v vm.VMAttachCredentialsSerialized, fullscreen bool) error {
	if err := nc.checkRemoteViewer(); err != nil {
		return err
	}

	p := params.VMAttachCreds{Pwd: v.Pwd}
	creds, err := nc.vmGetSpiceCreds(v.ID.String(), p)
	if err != nil {
		return err
	}

	err = nc.runRemoteViewer(creds.Name, creds.SpicePort, creds.SpicePwd, fullscreen)
	if err != nil {
		return err
	}

	return nil
}

func (nc *NexusClient) VMAttachFromPwd(pwd string, fullscreen bool) error {
	if err := nc.checkRemoteViewer(); err != nil {
		return err
	}

	p := params.VMAttachCreds{Pwd: pwd}
	creds, err := nc.vmGetAnySpiceCreds(p)
	if err != nil {
		return err
	} else {
		err := nc.runRemoteViewer(creds.Name, creds.SpicePort, creds.SpicePwd, fullscreen)
		if err != nil {
			return err
		}
	}

	return nil
}

func (nc *NexusClient) VMAddAccess(vmID, vmName, email string, vmAccessCaps *params.VMAddAccess) error {
	resp, err := nc.client.R().SetBody(vmAccessCaps).Put("/vms/" + vmID + "/access/" + email)
	if err != nil {
		return err
	}

	if resp.IsSuccess() {
		return nil
	} else {
		return response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) GetAttachVMsCreds() ([]vm.VMAttachCredentialsSerialized, error) {
	resp, err := nc.client.R().Get(attachVMs)
	if err != nil {
		return nil, err
	}

	if resp.IsSuccess() {
		vmCreds, err := deserializeVMsCredentials(resp)
		if err != nil {
			return nil, err
		}
		return vmCreds, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) GetDelVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(delVMs)
}

func (nc *NexusClient) GetEditAccessVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(editaccessVMs)
}

func (nc *NexusClient) GetEditVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(editVMs)
}

func (nc *NexusClient) GetExportDirVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(exportdirVMs)
}

func (nc *NexusClient) GetImportFileVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(importfileVMs)
}

func (nc *NexusClient) GetListVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(listVMs)
}

func (nc *NexusClient) GetRebootVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(rebootVMs)
}

func (nc *NexusClient) GetStartVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(startVMs)
}

func (nc *NexusClient) GetStopVMs() ([]vm.VMNetworkSerialized, error) {
	return nc.getVMs(stopVMs)
}

func (nc *NexusClient) GetListVM(vmID string) (*vm.VMNetworkSerialized, error) {
	resp, err := nc.client.R().Get("/vms/" + vmID)
	if err != nil {
		return nil, err
	}

	if resp.IsSuccess() {
		vm, err := deserializeVM(resp)
		if err != nil {
			return nil, err
		}
		return vm, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

//========================================================================================
// Private functions
//========================================================================================

func (nc *NexusClient) vmGetSpiceCreds(vmID string, p params.VMAttachCreds) (*vm.VMSpiceCredentialsSerialized, error) {
	resp, err := nc.client.R().SetBody(p).Post("/vms/" + vmID + "/spicecreds")
	if err != nil {
		return nil, err
	}

	return handleResponse(resp)
}

func (nc *NexusClient) vmGetAnySpiceCreds(p params.VMAttachCreds) (*vm.VMSpiceCredentialsSerialized, error) {
	resp, err := nc.client.R().SetBody(p).Post("/vms/spicecreds")
	if err != nil {
		return nil, err
	}

	return handleResponse(resp)
}

func (nc *NexusClient) getVMs(route string) ([]vm.VMNetworkSerialized, error) {
	resp, err := nc.client.R().Get(route)
	if err != nil {
		return nil, err
	}

	if resp.IsSuccess() {
		vms, err := deserializeVMs(resp)
		if err != nil {
			return nil, err
		}
		return vms, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

func (nc *NexusClient) vmIdToAttachCreds(vmID string) (*vm.VMAttachCredentialsSerialized, error) {
	resp, err := nc.client.R().Get("/vms/" + vmID + "/attach")
	if err != nil {
		return nil, err
	}

	if resp.IsSuccess() {
		vmCreds, err := deserializeVMCredentials(resp)
		if err != nil {
			return nil, err
		}
		return vmCreds, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

func handleResponse(resp *resty.Response) (*vm.VMSpiceCredentialsSerialized, error) {
	if resp.IsSuccess() {
		var creds vm.VMSpiceCredentialsSerialized
		if err := json.Unmarshal(resp.Body(), &creds); err != nil {
			return nil, errors.New("Failed deserializing VM spice credentials: " + err.Error())
		}
		if err := validator.New(validator.WithRequiredStructEnabled()).Struct(creds); err != nil {
			return nil, errors.New("Failed validating VM spice credentials: " + err.Error())
		}
		return &creds, nil
	} else {
		return nil, response.ErrorToMsg(resp)
	}
}

// Deserialize a list of VMs from an http response.
func deserializeVMs(resp *resty.Response) ([]vm.VMNetworkSerialized, error) {
	vms := []vm.VMNetworkSerialized{}
	if err := json.Unmarshal(resp.Body(), &vms); err != nil {
		return nil, err
	}
	return vms, nil
}

// Deserialize a single VM from an http response.
func deserializeVM(resp *resty.Response) (*vm.VMNetworkSerialized, error) {
	vm := vm.VMNetworkSerialized{}
	if err := json.Unmarshal(resp.Body(), &vm); err != nil {
		return nil, err
	}
	if err := validator.New(validator.WithRequiredStructEnabled()).Struct(vm); err != nil {
		return nil, err
	}
	return &vm, nil
}

// Deserialize a list of VM credentials from an http response.
func deserializeVMsCredentials(resp *resty.Response) ([]vm.VMAttachCredentialsSerialized, error) {
	vmsCreds := []vm.VMAttachCredentialsSerialized{}
	if err := json.Unmarshal(resp.Body(), &vmsCreds); err != nil {
		return nil, err
	}
	return vmsCreds, nil
}

// Deserialize a VM credentials from an http response.
func deserializeVMCredentials(resp *resty.Response) (*vm.VMAttachCredentialsSerialized, error) {
	var vmCreds vm.VMAttachCredentialsSerialized
	if err := json.Unmarshal(resp.Body(), &vmCreds); err != nil {
		return nil, err
	}
	if err := validator.New(validator.WithRequiredStructEnabled()).Struct(vmCreds); err != nil {
		return nil, err
	}
	return &vmCreds, nil
}
